#include "Halide.h"
#include "halide_benchmark.h"
#include <cstdio>

using namespace Halide;
using namespace Halide::Tools;

void simple_version(float *A, float *B, float *C, int width, int stride) {
    for (int iy = 0; iy < width; iy++) {
        for (int ix = 0; ix < width; ix++) {
            float *cc = C + iy * stride + ix;
            *cc = 0.0f;

            for (int ik = 0; ik < width; ik++) {
                *cc = *cc + A[iy * stride + ik] * B[ik * stride + ix];
            }
        }
    }
}

int main(int argc, char **argv) {

    const int matrix_size = 992;

    ImageParam A(type_of<float>(), 2);
    ImageParam B(type_of<float>(), 2);
    std::vector<Argument> args(2);
    args[0] = A;
    args[1] = B;

    Var x("x"), xi("xi"), xo("xo"), y("y"), yo("yo"), yi("yi"), yii("yii"), xii("xii");
    Func matrix_mul("matrix_mul");

    RDom k(0, matrix_size);
    RVar ki;

    matrix_mul(x, y) += A(k, y) * B(x, k);

    Func out;
    out(x, y) = matrix_mul(x, y);

    int vlen, arch_bits;
    arch_bits = 32;
    vlen = 256;

    //ISSUE: no llvm error if we replace y with x in the following line:
    out.vectorize(y, vlen/arch_bits);

    // matrix_mul.compute_at(out, y)
    //     .vectorize(x, vlen/arch_bits);

    // matrix_mul.update(0)
    //     .vectorize(x, vlen/arch_bits);

    out
        .bound(x, 0, matrix_size)
        .bound(y, 0, matrix_size);

    // out.compile_jit();

    Buffer<float> mat_A(matrix_size, matrix_size);
    Buffer<float> mat_B(matrix_size, matrix_size);
    Buffer<float> output(matrix_size, matrix_size);

    // init randomly
    for (int iy = 0; iy < matrix_size; iy++) {
        for (int ix = 0; ix < matrix_size; ix++) {
            mat_A(ix, iy) = (rand() % 256) / 256.0f;
            mat_B(ix, iy) = (rand() % 256) / 256.0f;
        }
    }

    A.set(mat_A);
    B.set(mat_B);

    out.realize(output);
    
    out.print_loop_nest();

    // Uncomment to see the generated assembly.
    {
        // printf("inferring mat mul args \n");
        // std::vector<Argument> args = matrix_mul.infer_arguments();
        printf("defining SVE target \n");
        Target target("host-no_asserts-no_runtime-no_bounds_query");
        target.os = Target::Linux;                // The operating system
        target.arch = Target::ARM;                  // The CPU architecture
        target.bits = arch_bits;                           // The bit-width of the architecture
        target.vector_bits = vlen;                           
        std::vector<Target::Feature> rvv_features;  // A list of features to set
        rvv_features.push_back(Target::SVE);
        target.set_features(rvv_features);
        // out.compile_to_lowered_stmt("matrix_multiplication_pseudo.c", args, Text, target);
        printf("compiling to sve \n");
        out.compile_to_assembly("matrix_multiplication.sve.S", args, "out", target);
        printf("completed compile to sve \n");
    }
        // RVV generated assembly.
    {
        // printf("inferring mat mul args \n");
        // std::vector<Argument> args = matrix_mul.infer_arguments();
        printf("defining RVV target \n");
        Target target("host-no_asserts-no_runtime-no_bounds_query");
        target.os = Target::Linux;                // The operating system
        target.arch = Target::RISCV;                  // The CPU architecture
        target.bits = 64;                           // The bit-width of the architecture
        target.vector_bits = 128;                           
        std::vector<Target::Feature> rvv_features;  // A list of features to set
        rvv_features.push_back(Target::RVV);
        target.set_features(rvv_features);
        // out.compile_to_lowered_stmt("matrix_multiplication_pseudo.c", args, Text, target);
        printf("compiling to rvv \n");
        out.compile_to_assembly("matrix_multiplication.rvv.S", args, "out", target);
        printf("completed compile to rvv \n");
    }
   
    // float gflops = 2.0f * matrix_size * matrix_size * matrix_size / 1e9f;

    // printf("Halide: %fms, %f GFLOP/s\n\n", t * 1e3, (gflops / t));

    // double t = benchmark([&]() {
    //     out.realize(output);
    // });

    // check results
    Buffer<float> output_ref(matrix_size, matrix_size);
    Buffer<float> output_halide(matrix_size, matrix_size);

    simple_version(mat_A.data(), mat_B.data(), output_ref.data(), mat_A.width(), mat_A.stride(1));
    out.realize(output_halide);

    bool halide_correct = true;
    for (int iy = 0; iy < matrix_size && halide_correct; iy++) {
        for (int ix = 0; ix < matrix_size; ix++) {
            halide_correct = halide_correct && (std::abs(output_ref(ix, iy) - output_halide(ix, iy)) < 0.001f);
        }
    }

    if (halide_correct) {
        printf("Halide results - OK\n");
    } else {
        printf("Halide results - FAIL\n");
        return 1;
    }

    printf("Success!\n");
    return 0;
}
